/* 
 * Argyll Color Correction System
 *
 * Abstract instrument class implemenation
 *
 * Author: Graeme W. Gill
 * Date:   10/3/2001
 *
 * Copyright 2001 Graeme W. Gill
 * All rights reserved.
 *
 * This material is licenced under the GNU GENERAL PUBLIC LICENCE :-
 * see the Licence.txt file for licencing details.
 */

/* 
   If you make use of the instrument driver code here, please note
   that it is the author(s) of the code who take responsibility
   for its operation. Any problems or queries regarding driving
   instruments with the Argyll drivers, should be directed to
   the Argyll's author(s), and not to any other party.

   If there is some instrument feature or function that you
   would like supported here, it is recommended that you
   contact Argyll's author(s) first, rather than attempt to
   modify the software yourself, if you don't have firm knowledge
   of the instrument communicate protocols. There is a chance
   that an instrument could be damaged by an incautious command
   sequence, and the instrument companies generally cannot and
   will not support developers that they have not qualified
   and agreed to support.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include "copyright.h"
#include "config.h"
#include "xspect.h"
#include "insttypes.h"
#include "icoms.h"
#include "usbio.h"
#include "inst.h"
#include "dtp41.h"
#include "dtp51.h"
#include "dtp92.h"
#include "gretag.h"
#include "ss.h"

/* inst error code interpretation */
static char *interp_code(struct _inst *p, inst_code ev) {

	switch(ev & inst_mask) {
		case inst_ok:
			return "No error";
		case inst_notify:
			return "Notification";
		case inst_warning:
			return "Warning";
		case inst_internal_error:
			return "Internal software error";
		case inst_coms_fail:
			return "Communications failure";
		case inst_unknown_model:
			return "Not expected instrument model";
		case inst_protocol_error:
			return "Communication protocol breakdown";
		case inst_user_abort:
			return "User hit Escape";
		case inst_user_term:
			return "User hit Terminate Key";
		case inst_misread:
			return "Measurement misread";
		case inst_needs_cal:
			return "Instrument needs calibration";
		case inst_unsupported:
			return "Unsupported function";
		case inst_unexpected_reply:
			return "Unexpected Reply";
		case inst_wrong_config:
			return "Wrong Configuration";
		case inst_hardware_fail:
			return "Hardware Failure";
		case inst_other_error:
			return "Non-specific error";
	}
	return "Unknown inst error code";

}

/* Default instrument user interaction callback. */
/* This will use stdio. */
static int inst_user(
	inst *p,
	inst_uiact it,		/* Interaction type */
	int nm,				/* Number of messages */
	int nch,			/* Number of choices to accept */
    ...					/* sequence of pointers to strings */
) {
	va_list args;
	int i, rv = 0;
	char c;

	if (nch > 35)
		nch = 35;

	if (nch > nm)
		nch = nm;

	for (;;) {
		va_start(args, nch);
		for (i = 0; i < nm; i++) {
			char *mes = va_arg(args, char *);
			if ((nm-i) <= nch) {
				int j = i-nm+nch+1;
				if (j < 10)
					c = '0' + j;
				else
					c = 'A' + j - 10; 
				printf(" %c: %s",c,mes);
			} else {
				printf("%s",mes);
			}
			
		}
		va_end(args);
		if (nch <= 0)
			return rv;

		fflush(stdout);
		empty_con_chars();
		c = next_con_char();
		printf("\n");
		if (c >= '0' && c <= '9')
			rv = c - '0';
		else if (c >= 'A' && c <= 'Z')
			rv = c - 'A' + 10;
		else if (c >= 'a' && c <= 'z')
			rv = c - 'a' + 10;
		else
			rv = 0;
		if (rv >= 1 && rv <= nch)
			return rv;
	}
}

/* Virtual constructor */
/* Return NULL for unknown instrument */
/* Can hand in instUnknown for a USB port */
extern inst *new_inst(
int comport,
instType itype) {
	instType atype;		/* Actual type */
	icoms *icom;		/* Optional icoms to hand in */
	inst *p;

	if ((icom = new_icoms()) == NULL)
		return NULL;

	/* Set instrument type from USB port, if not specified */
	atype = usb_is_usb_portno(icom, comport);
	if (itype == instUnknown && atype != instUnknown) {
		itype = atype;
	} else if (itype != instUnknown && atype == instUnknown) {
		atype = itype;
	}

	/* itype and atype will be the same now */
	if (itype == instDTP41 && atype == instDTP41)
		p = (inst *)new_dtp41(icom);
	else if (itype == instDTP51 && atype == instDTP51)
		p = (inst *)new_dtp51(icom);
	else if ((itype == instDTP92 && atype == instDTP92) || 
	         (itype == instDTP94 && atype == instDTP94))
		p = (inst *)new_dtp92(icom);
	else if ((itype == instSpectrolino  && atype == instSpectrolino) ||
			 (itype == instSpectroScan  && atype == instSpectroScan) ||
			 (itype == instSpectroScanT && atype == instSpectroScanT))
		p = (inst *)new_gretag(icom);
/* NYI
	else if (itype == instSpectrocam && atype == instSpectrocam)
		p = (inst *)new_spc(icom);
*/
	else
		return NULL;

	p->inst_interp_error = interp_code;

	/* Provide default user interaction callback */
	if (p->user == NULL)
		p->user = inst_user;

	return p;
}



