
/* 
 * Argyll Color Correction System
 *
 * General USB I/O support
 *
 * Author: Graeme W. Gill
 * Date:   2006/22/4
 *
 * Copyright 2006 Graeme W. Gill
 * All rights reserved.
 *
 * This material is licenced under the GNU GENERAL PUBLIC LICENCE :-
 * see the Licence.txt file for licencing details.
 */

/* These routines supliement the class code in ntio.c and unixio.c */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#if defined(UNIX)
#include <termios.h>
#endif
#include "copyright.h"
#include "config.h"
#include "numsup.h"
#include "xspect.h"
#include "insttypes.h"
#include "icoms.h"
#include "usbio.h"

#include <usb.h>

#define DEBUG

#ifdef ENABLE_USB

/* Check a USB Vendor and product ID, and add the device */
/* to the icoms path if it is supported. Return nz if it was added. */
static int usb_check_and_add(
struct _icoms *p,
struct usb_device *dev
) {
	instType itype;

	if ((itype = inst_usb_match(dev->descriptor.idVendor, dev->descriptor.idProduct))
	                                                                != instUnknown) {
		char pname[400];

		/* Create a path/identification */
		/* (devnum doesn't seem valid ?) */
#if defined(UNIX)
			sprintf(pname,"usb:/bus%d/dev%d (%s)",dev->bus->location >> 24, dev->devnum, inst_name(itype));
#else
			sprintf(pname,"usb:/bus%lu/dev%d (%s)",dev->bus->location, dev->devnum, inst_name(itype));
#endif

		/* Add the path to the list */
		if (p->paths == NULL) {
			if ((p->paths = (icompath **)calloc(sizeof(icompath *), 1 + 1)) == NULL)
				error("icoms: calloc failed!");
		} else {
			if ((p->paths = (icompath **)realloc(p->paths,
			                     sizeof(icompath *) * (p->npaths + 2))) == NULL)
				error("icoms: realloc failed!");
			p->paths[p->npaths+1] = NULL;
		}
		if ((p->paths[p->npaths] = malloc(sizeof(icompath))) == NULL)
			error("icoms: malloc failed!");
		p->paths[p->npaths]->dev = dev;
		p->paths[p->npaths]->itype = itype;
		if ((p->paths[p->npaths]->path = strdup(pname)) == NULL)
			error("icoms: strdup failed!");
		p->npaths++;
		p->paths[p->npaths] = NULL;
		return 1;
	}

	return 0;
}

#endif /* ENABLE_USB */

/* Add paths to USB connected instruments, to the existing */
/* icompath paths in the icoms structure. */
void usb_get_paths(
struct _icoms *p 
) {
#ifdef ENABLE_USB
	struct usb_bus *bus;

	/* Scan the USB busses for instruments we recognise */
	/* We're not expecting any of our unstruments to be an interface on a device. */
   	usb_init();
   	usb_find_busses();
   	usb_find_devices();
    
   	for (bus = usb_get_busses(); bus != NULL; bus = bus->next) {
		struct usb_device *dev;
   		for (dev = bus->devices; dev != NULL; dev = dev->next) {
			usb_check_and_add(p, dev);
		}
   	}
#endif /* ENABLE_USB */
}


/* Return the instrument type if the port number is USB, */
/* and instUnknown if it is not. */
instType usb_is_usb_portno(
	icoms *p, 
	int port		/* Enumerated port number, 1..n */
) {
	
	if (p->paths == NULL)
		icoms_get_paths(p);

	if (port <= 0 || port > p->npaths)
		error("icoms - set_ser_port: port number out of range!");

#ifdef ENABLE_USB
	if (p->paths[port-1]->dev != NULL)
		return p->paths[port-1]->itype;
#endif /* ENABLE_USB */

	return instUnknown;
}



/*  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
#ifdef ENABLE_USB

/* Static list so that all open USB connections can be closed on a SIGKILL */
static icoms *icoms_list = NULL;

/* Flag set when we're in a USB read or write */
static int in_usb_rw = 0;

/* Clean up any open USB ports and exit */
static void icoms_cleanupandexit() {
	icoms *pp, *np;
//printf("~1 icoms_cleanipandexit invoked\n");

#if defined(UNIX)
	/* This is a bit of a hack to compensate for the fact */
	/* that a ^C will kill the program while ICANON is off. */
	/* It's really better to restore the original attributes, */
	/* even when USB is not compiled in. */
	struct termios news;
	if (tcgetattr(STDIN_FILENO, &news) >= 0) {
		news.c_lflag |= (ICANON | ECHO);
		tcsetattr(STDIN_FILENO,TCSANOW, &news);
	}
#endif

	for (pp = icoms_list; pp != NULL; pp = np) { 
		np = pp->next;
//printf("~1 closing usb port 0x%x\n",pp);
		/* There's a problem here if have more than one USB port */
		/* open - win32 doesn't return from the system call. */
		/* Should we depend on usb read/write routines to call cleanup ? */
		usb_close_port(pp);
	}
	exit(0);
}

/* On something killing our process, deal with USB cleanup */
static void icoms_sighandler(int arg) {
//printf("~1 signal handler invoked\n");
	if (in_usb_rw != 0)
		in_usb_rw = 2;
	icoms_cleanupandexit();
}

/* Our versions of usblib read/write, that exit if a signal was caught */
/* This is so that MSWindows works properly */
int icoms_usb_interrupt_write(usb_dev_handle *dev, int ep, char *bytes, int size, int timeout) {
	int rv;

	in_usb_rw = 1;
	rv = usb_interrupt_write(dev, ep, bytes, size, timeout);
	if (in_usb_rw == 2)
		exit(0);

	in_usb_rw = 0;
	return rv;
}

int icoms_usb_interrupt_read(usb_dev_handle *dev, int ep, char *bytes, int size, int timeout) {
	int rv;

	in_usb_rw = 1;
	rv = usb_interrupt_read(dev, ep, bytes, size, timeout);
	if (in_usb_rw == 2)
		exit(0);

	in_usb_rw = 0;
	return rv;
}

#endif /* ENABLE_USB */
/*  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */



/* Close an open USB port */
/* If we don't do this, the port and/or the device may be left in an unusable state. */
void usb_close_port(icoms *p) {

#ifdef ENABLE_USB
	if (p->debug) fprintf(stderr,"usb_close_port() called\n");

	if (p->usbh != NULL) {
		usb_reset(p->usbh);
		usb_release_interface(p->usbh, p->ifce);
		usb_close(p->usbh);

		if (p->debug) fprintf(stderr,"usb port has been released and closed\n");
	}

	/* Find it and delete it from our static cleanup list */
	if (icoms_list != NULL) {
		if (icoms_list == p) {
			icoms_list = p->next;
			if (icoms_list == NULL) {
//printf("~1 removing signal handler\n");
#if defined(UNIX)
				signal(SIGHUP, SIG_DFL);
#endif /* UNIX */
				signal(SIGINT, SIG_DFL);
				signal(SIGTERM, SIG_DFL);
			}
		} else {
			icoms *pp;
			for (pp = icoms_list; pp != NULL; pp = pp->next) { 
				if (pp->next == p) {
					pp->next = p->next;
					break;
				}
			}
		}
	}
#endif /* ENABLE_USB */
}

/* Declaration of needed function in ntio.c or unixio.c */
void icoms_close_port(icoms *p);

/* Open a USB port */
void usb_open_port(
icoms *p,
int    port,		/* USB com port, 1 - N, 0 for no change. */
int    config,	/* Configuration */
int    iface,		/* Interface */
int    wr_ep,		/* Write end point */
int    rd_ep,		/* Read end point */
int    rd_qa	/* Read quanta size */
) {
	if (p->debug) fprintf(stderr,"icoms: About to open the USB port\n");

	if (port >= 1) {
		if (p->is_open && port != p->port) {	/* If port number changes */
			icoms_close_port(p);
		}
	}

	/* Make sure the port is open */
	if (!p->is_open) {
		if (p->debug) fprintf(stderr,"icoms: USB port needs opening\n");

		if (p->ppath != NULL) {
			if (p->ppath->path != NULL)
				free(p->ppath->path);
			free(p->ppath);
		}

		if (p->paths == NULL)
			icoms_get_paths(p);

		if (port <= 0 || port > p->npaths)
			error("icoms - usb_open_port: port number out of range!");

		if (p->paths[port-1]->dev == NULL)
			error("icoms - usb_open_port: Not a USB port!");

		if ((p->ppath = malloc(sizeof(icompath))) == NULL)
			error("malloc() failed on com port path");
		*p->ppath = *p->paths[port-1];				/* Structure copy */
		if ((p->ppath->path = strdup(p->paths[port-1]->path)) == NULL)
			error("strdup() failed on com port path");
		p->port = port;

		if (p->debug) fprintf(stderr,"icoms: About to open USB port '%s'\n",p->ppath->path);

		if ((p->usbh = usb_open(p->ppath->dev)) == NULL)
			error("Opening USB port '%s' failed",p->ppath->path);

		p->cnfg = config;
		p->ifce = iface;
		p->wr_ep = wr_ep;
		p->rd_ep = rd_ep;
		p->rd_qa = rd_qa;
		/* (Should use bConfigurationValue ?) */
		if (usb_set_configuration(p->usbh, p->cnfg) < 0)
			error("Configuring USB port '%s' to %d failed",p->ppath->path,config);

		/* We're hardwired interface to 0 for the moment */
		/* (Second parameter is bInterfaceNumber) */
		if (usb_claim_interface(p->usbh, p->ifce) < 0)
			error("Claiming USB port '%s' interface %d failed",p->ppath->path,iface);

		p->is_usb = 1;
		p->is_open = 1;
		if (p->debug) fprintf(stderr,"icoms: USB port is now open\n");
	}

	if (p->debug) fprintf(stderr,"icoms: Clearing any USB errors\n");

	/* The interface seems to hang at times. Clearing halt */
	/* seems to unstick it. */
	usb_clear_halt(p->usbh, p->rd_ep);
	usb_clear_halt(p->usbh, p->wr_ep);

	if (icoms_list == NULL) {
//printf("~1 installing signal handler\n");
#if defined(UNIX)
		signal(SIGHUP, icoms_sighandler);
#endif /* UNIX */
		signal(SIGINT, icoms_sighandler);
		signal(SIGTERM, icoms_sighandler);
	}

	/* Add it to our static list, to allow automatic cleanup */
	p->next = icoms_list;
	icoms_list = p;

}

